/*
 * pcodevm
 * interpreter for the UCSD p-code byte machine as used with UCSD Pascal II.1,
 * UCSD Pascal Compiler II.0, including temporary I.5 translations
 *
 * Copyright GPL (c)2009 Oskar Schirmer
 *
 * Mein besonderer Dank gilt Peter Vortisch, der mir seinerzeit häufig
 * das richtige Buch zum Thema gegeben hat, so auch das Handbuch "Apple
 * Pascal Reference Manual", ohne welches dieses Programm zu schreiben
 * mir kaum möglich gewesen wäre.
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <getopt.h>
#include <math.h>
#include <errno.h>
#include <ncurses.h>
#include <signal.h>
#include <sys/poll.h>

#ifdef PATH_MAX
#define MAX_PATH PATH_MAX
#else
#define MAX_PATH 4096
#endif

#define DEBUG 1
#if DEBUG
#define deb(s...) ( { if (opt_debug) fprintf(stderr, s); } )
#else
#define deb(s...)
#endif

typedef uint8_t u8;
typedef int8_t s8;
typedef uint16_t u16;
typedef int16_t s16;
typedef uint32_t u32;
typedef int32_t s32;
typedef float real;

int opt_debug = 0;

#define BLOCKSIZE 512
#define NBSLOT 16
#define NBSEG 32
#define NBUNIT 13
#define PNIL 0

#define SETCRT 17
#define SETDIR 18
#define SETSEG 19
#define SETIOR 20
#define MAX_BANK 255
#if ((MAX_BANK < 1) || (MAX_BANK > 255))
#error "MAX_BANK must be in range 1..255"
#endif

#define unitblkd(u) \
        (((1 << 4 | 1 << 5 | 1 << 9 | 1 << 10 | 1 << 11 | 1 << 12) >> (u)) & 1)
static int unitr[NBUNIT];
static int unitw[NBUNIT];
static s8 unitt[NBUNIT];
static u8 apple;

static void exiterr(int err)
{
  if (abs(unitt[1]) == 2) {
    endwin();
  }
  exit(err);
}

#define warning(s...) ( { fprintf(stderr, "\n" s); fputs("\n", stderr); } )
#define execerr(s...) ( { warning(s); exiterr(1); } )

static char ctrlc = 0;
static void signalhandler(int sig)
{
  if (++ctrlc == 3) {
    exiterr(0);
  }
}

static void usage(char *arg0)
{
  fprintf(stderr,
    "Usage: %s [OPTION]... CODEFILE [CODELIB...]\n"
    "Execute pCode executable\n"
    "\n"
    "  -d, --disk <image>\tmount disk image"
        " (unit #4,5,9,10,11,12 in this order)\n"
    "  -l, --printer <file>\twrite printer output to file (unit #6)\n"
    "  -r, --rx <file>\tread serial input from file (unit #7)\n"
    "  -t, --tx <file>\twrite serial output to file (unit #8)\n"
    "  -T, --term\t\tuse curses and signal for CRT info and ctrl\n"
    "  -A, --apple\t\thandle APPLE2 hardware specific functions\n"
    "  -h, --help\t\tshow this help\n"
    "  -V, --version\t\tshow the version (" PCODEVM_VERSION ")\n"
    "  -D, --debug\t\tlots of debug information to stderr\n"
    "\n"
    "usually executes proc#1/seg#1 of the given CODEFILE,\n"
    "with some initialisation prior to it; always needs some global\n"
    "initialisation code and a system disk image, e.g. run a single program:\n"
    "  ./pcodevm PROGRAM.CODE SYSTEM.PASCAL --disk apple1.img\n"
    "or to run the operating system itself, e.g.:\n"
    "  ./pcodevm SYSTEM.PASCAL --disk apple1.img --term\n"
    "use ctrl-@ or triple ctrl-c to break when in --term mode.\n"
    , arg0);
}

/* check unix commandline options
 * returns the index into argv for the first non option argument
 * the return value is always less than argc
 */
static int chkoptions(int argc, char *argv[])
{
  static const struct option longopt[] = {
    {"disk", 1, NULL, 'd'},
    {"printer", 1, NULL, 'l'},
    {"rx", 1, NULL, 'r'},
    {"tx", 1, NULL, 't'},
    {"term", 0, NULL, 'T'},
    {"apple", 0, NULL, 'A'},
    {"help", 0, NULL, 'h'},
    {"version", 0, NULL, 'V'},
    {"debug", 0, NULL, 'D'},
    {0, 0, 0, 0}
  };
  int i, f, d = 0;
  int optindex;
  char *endptr;
  memset(&unitr[0], -1, sizeof(unitr));
  memset(&unitw[0], -1, sizeof(unitw));
  memset(&unitt[0], 0, sizeof(unitt));
  unitr[1] = 0;
  unitw[1] = 1;
  unitr[2] = 0;
  unitt[1] = 1;
  unitt[2] = 1;
  unitt[6] = 1;
  unitt[7] = 1;
  unitt[8] = 1;
  apple = 0;
  while ((i = getopt_long(argc, argv, "d:l:r:t:TAhVD",
        longopt, &optindex)) >= 0) {
    switch (i) {
    case 'd':
      while (!unitblkd(++d)) {
        if (d >= NBUNIT) {
          i = 0;
          while (--d) i += unitblkd(d) ? 1 : 0;
          fprintf(stderr, "not enough disk slots: max=%d\n", i);
          usage(argv[0]);
          exit(1);
        }
      }
      f = open(optarg, O_RDWR);
      if (f < 0) f = open(optarg, O_RDONLY);
      else unitw[d] = f;
      if (f < 0) execerr("cannot open: %s", optarg);
      else unitr[d] = f;
      break;
    case 'l':
      i = 6; if (0) {
    case 't':
      i = 8; }
      if (unitw[i] >= 0) {
        fprintf(stderr, "unit assigned twice: #%d\n", i);
        usage(argv[0]);
        exit(1);
      }
      f = open(optarg, O_WRONLY|O_CREAT,
        S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH);
      if (f < 0) execerr("cannot open: %s", optarg);
      else unitw[i] = f;
      break;
    case 'r':
      i = 7;
      if (unitr[i] >= 0) {
        fprintf(stderr, "unit assigned twice: #%d\n", i);
        usage(argv[0]);
        exit(1);
      }
      f = open(optarg, O_RDONLY);
      if (f < 0) execerr("cannot open: %s", optarg);
      else unitr[i] = f;
      break;
    case 'T':
      unitt[1] = 2;
      unitt[2] = 2;
      break;
    case 'A':
      apple = 1;
      break;
    case 'V':
      fputs(PCODEVM_VERSION "\n", stderr);
      exit(0);
      break;
    case 'h':
      usage(argv[0]);
      exit(0);
      break;
    case 'D':
      opt_debug = 1;
      break;
    default:
      usage(argv[0]);
      exit(1);
      break;
    }
  }
  if (optind >= argc) {
    usage(argv[0]);
    exit(1);
  }
  return optind;
}

static u8 *ioresult;
static u8 *crtctrl;
static u32 segtable;
static u32 globdirp;

static int pcerrno(int e)
{
  switch (e) {
  case 0:       return 0;
  case EPERM:   return 16;
  case ENOENT:  return 10;
  case EIO:     return 1;
  case ENXIO:   return 2;
  case EBADF:   return 6;
  case EAGAIN:  return 4;
  case ENOMEM:  return 8;
  case EACCES:  return 16;
  case EFAULT:  return 9;
  case ENOTBLK: return 17;
  case EBUSY:   return 4;
  case EEXIST:  return 11;
  case ENODEV:  return 9;
  case ENFILE:  return 8;
  case EMFILE:  return 8;
  case ETXTBSY: return 4;
  case EFBIG:   return 8;
  case ENOSPC:  return 8;
  case EROFS:   return 16;
  case ENAMETOOLONG: return 7;
  default:      return 3;
  }
}

enum tsegkind {
  linked,
  hostseg,
  segproc,
  unitseg,
  seprtseg,
  unlinkedintrins,
  linkedintrins,
  dataseg
};

typedef struct procdescr {
  u8 *ienter;
  u8 *iexit;
  u8 *jtab;
  unsigned psize, dsize;
  int lexlev;
} procdescr;

typedef struct segdescr {
  u8 *code;
  unsigned codeleng;
  unsigned char segname[8];
  enum tsegkind segkind;
  unsigned proccnt;
  procdescr *proc;
  signed use;
} segdescr;

static segdescr seg[NBSEG] = {{0}};

const static u8 code41[] = {
  165, 1,       /* SYS_COM := */
  12, 152,
  158, 1,       /* allocate(288) */
  232,
  158, SETIOR,  /* set ioresult */
  232,
  162, 4,
  158, SETDIR,  /* global dirp */
  232,
  162, 31,
  158, SETCRT,  /* crt ctrl */
  232,          /* SYS_COM^ */
  162, 2,       /* .SYS_UNIT */
  4, 154,       /* := 4 */
  205, 4, 1,
  218,          /* OUTPUT */
  205, 0, 22,   /* writeln */
  159, 159,     /* INPUT=nil, OUTPUT=nil */
  205, 1, 1,
  214
};

const static u8 code40[] = {
  165, 1,       /* SYS_COM := */
  12, 152,
  158, 1,       /* allocate(288) */
  232,
  158, SETIOR,  /* set ioresult */
  232,
  162, 4,
  158, SETDIR,  /* global dirp */
  232,
  162, 31,
  158, SETCRT,  /* crt ctrl */
  232,
  162, 48,
  158, SETSEG,  /* locate segment table */
  232,          /* SYS_COM^ */
  162, 2,       /* .SYS_UNIT */
  4, 154,       /* := 4 */
  159, 204, 54, /* EMPTY_HEAP :=nil */
  1, 158, 38,   /* unitclear (1) */
  205, 4, 1,    /* INITIALIZE; */
  205, 0, 48,   /* Command; */
  202, 54,      /* until EMPTY_HEAP */
  159, 212, 254,/* = nil */
  214,
  12, 0
};

static u8 *mem;
static u8 *bank;
static u32 tos;
static u32 toh;

static u16 getle16(u8 *b)
{
  return (b[1] << 8) + b[0];
}

static void putle16(u8 *b, u16 v)
{
  b[0] = (u8)v;
  b[1] = (u8)(v >> 8);
}

static u32 align(u32 v)
{
  return v & ~1;
}

static void move_mem(u32 src, u32 dst, u32 count)
{
  memcpy(&mem[dst], &mem[src], count);
  if ((src | dst) & 1)
    memset(&bank[dst >> 1], 0, (count+1 + (dst & 1)) >> 1);
  else
    memcpy(&bank[dst >> 1], &bank[src >> 1], (count+1) >> 1);
}

static void fill_mem(u32 dst, u32 count, u8 value)
{
  memset(&mem[dst], value, count);
  memset(&bank[dst >> 1], 0, (count+1 + (dst & 1)) >> 1);
}

static void move_left(u32 src, u32 dst, u32 count)
{
  u32 s = src;
  u32 d = dst;
  u32 c;
  for (c = count; c > 0; c--)
    mem[d++] = mem[s++];
  if ((src | dst) & 1) {
    memset(&bank[dst >> 1], 0, (count+1 + (dst & 1)) >> 1);
  } else {
    s = src >> 1;
    d = dst >> 1;
    for (c = (count+1) >> 1; c > 0; c -= 1)
      bank[d++] = bank[s++];
  }
}

static void move_right(u32 src, u32 dst, u32 count)
{
  u32 s = src + count;
  u32 d = dst + count;
  u32 c;
  for (c = count; c > 0; c--)
    mem[--d] = mem[--s];
  if ((src | dst) & 1) {
    memset(&bank[dst >> 1], 0, (count+1 + (dst & 1)) >> 1);
  } else {
    s = (src >> 1) + ((count+1) >> 1);
    d = (dst >> 1) + ((count+1) >> 1);
    for (c = (count+1) >> 1; c > 0; c -= 1)
      bank[--d] = bank[--s];
  }
}

static u8 *relpos16(u8 *l)
{
  return l - getle16(l);
}

static void check_mem(void)
{
  if ((s32)toh > (s32)tos)
    execerr("stack overflow");
}

static void check_stack(unsigned size)
{
  if ((tos + size) > ((MAX_BANK + 1) << 16))
    execerr("stack underflow");
}

static void push_stack(unsigned size)
{
  tos -= size;
}

static void pop_stack(unsigned size)
{
  tos += size;
}

static int is_ref(u32 index)
{
  int r;
  r = bank[index >> 1];
  if (r && (index & 1)) execerr("unaligned access");
  return r;
}

static void check_int(u32 index)
{
  if (is_ref(index)) execerr("type mismatch (int)");
}

static void check_ref(u32 index)
{
  if (!is_ref(index)) execerr("type mismatch (ref)");
}

static void check_align(u32 index)
{
  if (index & 1) execerr("alignment fault");
}

static void store_any(u32 index, u32 value)
{
  *((u16*)&mem[index]) = (u16)value;
  bank[index >> 1] = (u8)(value >> 16);
}

static void store_any_check(u32 index, u32 value)
{
  check_align(index);
  store_any(index, value);
}

static u32 load_any(u32 index)
{
  return ((u32)bank[index >> 1] << 16) + *(u16*)&mem[index];
}

static void store_int(u32 index, u16 value)
{
  *((u16*)&mem[index]) = value;
  bank[index >> 1] = 0;
}

static void store_int_check(u32 index, u16 value)
{
  check_align(index);
  store_int(index, value);
}

static void push_int(u16 value)
{
  push_stack(2);
  store_int(tos, value);
}

static void push_int_check(u16 value)
{
  push_int(value);
  check_mem();
}

static u16 load_int(u32 index)
{
  return *(u16*)&mem[index];
}

static u16 load_int_check(u32 index)
{
  check_align(index);
  check_int(index);
  return load_int(index);
}

static u16 pop_int_check(void)
{
  u16 r;
  check_stack(2);
  check_int(tos);
  r = load_int(tos);
  pop_stack(2);
  return r;
}

static void push_any(u32 value)
{
  push_stack(2);
  store_any(tos, value);
}

static void push_any_check(u32 value)
{
  push_any(value);
  check_mem();
}

static store_ref(u32 index, u32 value)
{
  u16 h;
  *((u16*)&mem[index]) = (u16)value;
  h = value >> 16;
  if (!h || (h > MAX_BANK)) execerr("address out of range");
  bank[index >> 1] = (u8)h;
}

static store_ref_check(u32 index, u32 value)
{
  check_align(index);
  store_ref(index, value);
}

static void push_ref(u32 value)
{
  push_stack(2);
  store_ref(tos, value);
}

static void push_ref_check(u32 value)
{
  push_ref(value);
  check_mem();
}

static u32 load_ref_check(u32 index)
{
  check_align(index);
  check_ref(index);
  return load_any(index);
}

static u32 pop_ref_check(void)
{
  u32 r;
  check_stack(2);
  check_ref(tos);
  r = load_any(tos);
  pop_stack(2);
  return r;
}

static u32 pop_any(void)
{
  u32 r = load_any(tos);
  pop_stack(2);
  return r;
}

static u32 pop_any_check(void)
{
  check_stack(2);
  return pop_any();
}

union real_transform {
  real r;
  u16 w[2];
};

static void store_real(u32 index, real value)
{
  union real_transform x;
  x.r = value;
  store_int(index, x.w[0]);
  store_int(index + 2, x.w[1]);
}

static void push_real(real value)
{
  push_stack(4);
  store_real(tos, value);
}

static void push_real_check(real value)
{
  push_real(value);
  check_mem();
}

static real load_real_check(u32 index)
{
  union real_transform x;
  x.w[0] = load_int_check(index);
  check_int(index + 2);
  x.w[1] = load_int(index + 2);
  return x.r;
}

static real pop_real_check(void)
{
  real r;
  check_stack(4);
  r = load_real_check(tos);
  pop_stack(4);
  return r;
}

static u32 new_heap(unsigned size)
{
  u32 r = toh;
  toh += size;
  check_mem();
  if (globdirp)
    store_any_check(globdirp, 0);
  return r;
}

static u8 *readcode(unsigned char *name)
{
  u8 *d = NULL;
  int f, r;
  unsigned l, i;
  struct stat b;
  unsigned char t[MAX_PATH + 16];
  if (name != NULL) {
    f = open(name, O_RDONLY);
    if (f < 0) execerr("cannot open: %s", name);
    if (fstat(f, &b)) execerr("stat failed");
    l = b.st_size;
    if (l < BLOCKSIZE) execerr("short code file");
    d = &mem[new_heap(align(l+1))];
    i = 0;
    while (i < l) {
      r = read(f, &d[i], l-i);
      if (r <= 0) execerr("read error");
      i += r;
    }
    close(f);
    for (r = 0; r < NBSLOT; r++) {
      i = getle16(&d[4*r]) * BLOCKSIZE + getle16(&d[4*r+2]);
      if (i > l) execerr("segment exceeds code file");
    }
    deb("%7d %s\n", l, name);
  }
  return d;
}

static void unitread(u16 u, u32 a, s16 s, u16 b)
{
  *ioresult = pcerrno((u >= NBUNIT) ? ENXIO : (unitr[u] < 0) ? ENODEV : 0);
  if (!*ioresult) {
    int i;
    if (unitblkd(u)) {
      deb("seek(%d[%d], 512 * %d)\n", u, unitr[u], b);
      if (lseek(unitr[u], 512 * b, SEEK_SET) < 0) {
        *ioresult = pcerrno(errno);
        return;
      }
    }
    deb("read(%d[%d], %06lx, %d)\n", u, unitr[u], a, s);
    memset(&bank[a >> 1], 0, (s+1 + (a & 1)) >> 1);
    i = 0;
    while (i < s) {
      if (crtctrl && abs(unitt[u]) == 2) {
        int k, q;
        do {
          if (ctrlc) {
            k = crtctrl[29];
            ctrlc = 0;
          } else {
            k = getch();
          }
        } while (k == ERR);
        deb("input getch %d\n", k);
        switch (k) {
        case KEY_UP:    q = crtctrl[16]; break;
        case KEY_DOWN:  q = crtctrl[17]; break;
        case KEY_LEFT:  q = crtctrl[18]; break;
        case KEY_RIGHT: q = crtctrl[19]; break;
        case KEY_END:   q = crtctrl[20]; break;
        /* crtctrl[21]: FLUSH */
        /* crtctrl[22]: BREAK */
        /* crtctrl[23]: STOP */
        case KEY_BACKSPACE:
        case KEY_DC:    q = crtctrl[24]; break;
        case KEY_DL:    q = crtctrl[26]; break;
        case 27:        q = crtctrl[27]; break;
        case 10:        q = 13; break;
        default:        q = 0;
        }
        if (q) k = q;
        deb("input char code %d\n", k);
        mem[a+i] = k;
        i += 1;
      } else {
        int r = read(unitr[u], &mem[a+i], s - i);
        if (r < 0) {
          *ioresult = pcerrno(errno);
          return;
        }
        if (r == 0) break;
        if (unitt[u]) {
          do {
            if (mem[a+i] == 10) mem[a+i] = 13;
          } while (i++, --r > 0);
        } else {
          i += r;
        }
      }
    }
  }
}

static void unitwrite(u16 u, u32 a, s16 s, u16 b)
{
  *ioresult = pcerrno((u >= NBUNIT) ? ENXIO :
    (unitw[u] < 0) ? (unitr[u] < 0) ? ENODEV :
      EPERM : 0);
  if (!*ioresult) {
    int i;
    u8 *m;
    if (unitblkd(u)) {
      deb("seek(%d[%d], 512 * %d)\n", u, unitw[u], b);
      if (lseek(unitw[u], 512 * b, SEEK_SET) < 0) {
        *ioresult = pcerrno(errno);
        return;
      }
    }
    deb("write(%d[%d], %06lx, %d)\n", u, unitw[u], a, s);
    if (unitt[u]) {
      s8 d = unitt[u];
      u8 c;
      b = 0;
      for (i = 0; i < s; i++) {
        if (d < 0) {
          b += (u8)(mem[a+i] - 32);
          d = -d;
        } else if (mem[a+i] == 30) {
          b += 3;
          i += 2;
        } else if (mem[a+i] == 16) d = -d;
        else b += (mem[a+i] == 13) ? 2 : 1;
      }
      m = malloc(b);
      b = 0;
      c = 0;
      d = unitt[u];
      for (i = 0; i < s; i++) {
        if (d < 0) {
          c = mem[a+i] - 32;
          while (c > 0) {
            m[b++] = ' ';
            c -= 1;
          }
          d = -d;
        } else if (c > 0) {
          c -= 1;
          m[b++] = mem[a+i];
        } else if (mem[a+i] == 16) {
          d = -d;
        } else {
          m[b++] = mem[a+i];
          switch (mem[a+i]) {
          case 13:  m[b++] = 10;
                    break;
          case 30:  c = 2;
                    break;
          }
        }
      }
      unitt[u] = d;
      s = b;
    } else {
      m = &mem[a];
    }
    i = 0;
    if (crtctrl && abs(unitt[u]) == 2) {
      int h, w;
      getmaxyx(stdscr, h, w);
      putle16(&crtctrl[12], h);
      putle16(&crtctrl[14], w);
      while (i < s) {
        int x, y;
        switch (m[i]) {
        case 10:
          getyx(stdscr, y, x);
          if (y+1 >= h) scroll(stdscr);
          else move(y+1, x);
          break;
        case 13:
          getyx(stdscr, y, x);
          move(y, 0);
          break;
        case 30:
          x = m[++i] - 32;
          y = m[++i] - 32;
          move(y, x);
          break;
        default:
          if (crtctrl[1] == m[i]) {
            move(0, 0);
          } else if (crtctrl[2] == m[i]) {
            clrtobot();
          } else if (crtctrl[3] == m[i]) {
            clrtoeol();
          } else if (crtctrl[4] == m[i]) {
            getyx(stdscr, y, x);
            move(y, x+1);
          } else if (crtctrl[5] == m[i]) {
            getyx(stdscr, y, x);
            if (y) move(y-1, x);
          } else if (crtctrl[6] == m[i]) {
            getyx(stdscr, y, x);
            if (x) move(y, x-1);
          } else if (crtctrl[8] == m[i]) {
            getyx(stdscr, y, x);
            move(y, 0);
            clrtoeol();
            move(y, x);
          } else if (crtctrl[9] == m[i]) {
            erase();
          } else {
        case 0:
            printw("%c", m[i]);
          }
        }
        i += 1;
      }
      refresh();
    } else while (i < s) {
      int r = write(unitw[u], &m[i], s - i);
      if (r < 0) {
        *ioresult = pcerrno(errno);
        break;
      }
      if (r == 0) break;
      i += r;
    }
    if (unitt[u]) free(m);
  }
}

static void listsegment(unsigned s, u8 *c, unsigned l, signed u)
{
  unsigned p, q;
  p = c[l-1];
  seg[s].code = c;
  seg[s].codeleng = l;
  seg[s].proccnt = p;
  seg[s].use = u;
  deb("seg #%d, code %06x, len %04x, %d procs:\n",
      s, c - &mem[0], l, p);
  if (!seg[s].proc) {
    seg[s].proc = ((procdescr*)calloc(255, sizeof(procdescr))) - 1;
  }
  for (q = 1; q <= p; q++) {
    procdescr* pp = &seg[s].proc[q];
    u8 *j = relpos16(&c[l-2*(q+1)]);
    if (j[0] != q) warning("wrong proc ref (#%d: j0=%d)", q, j[0]);
    pp->jtab = j;
    pp->lexlev = (s8)j[1];
    pp->ienter = relpos16(j-2);
    pp->iexit = relpos16(j-4);
    pp->psize = getle16(j-6);
    pp->dsize = getle16(j-8);
    deb("  #%2d: lexl %2d, jtab %06lx, en/ex: %06lx/%06lx,"
        " par+dat:%4d+%d\n",
        q, pp->lexlev, j - mem, pp->ienter - mem, pp->iexit - mem,
        pp->psize, pp->dsize);
  }
}

static void fetchsegments(u8 *cf)
{
  if (cf) {
    unsigned i;
    for (i = 0; i < NBSLOT; i++) {
      unsigned l = getle16(&cf[4*i+2]);
      if (l > 0) {
        unsigned q, s;
        u8 *c;
        q = getle16(&cf[4*i]);
        c = &cf[q * BLOCKSIZE];
        s = c[l-2];
        if (seg[s].code) {
          deb("seg #%d: duplicate\n", s);
        } else {
          memcpy(&seg[s].segname[0], &cf[NBSLOT*4+8*i], 8);
          seg[s].segkind = getle16(&cf[NBSLOT*12+2*i]);
          deb("seg #%d, pos %04x, '%.8s',\n",
              s, q * BLOCKSIZE, &seg[s].segname[0]);
          listsegment(s, c, l, INT_MIN);
        }
      }
    }
  }
}

typedef struct activrec {
  u8 *ipc;
  unsigned seg;
  procdescr *proc;
  struct activrec *dyn;
  struct activrec *stat;
  struct activrec *base;
  u32 par;
} activrec;

static activrec *ar = NULL;
static activrec *unusedar = NULL;

static u8 *allocar(unsigned nseg, unsigned nproc, procdescr *p)
{
  u32 t;
  u8 *e;
  activrec *a = ar;
  activrec *n = unusedar;
  if (n) unusedar = n->dyn;
    else n = malloc(sizeof(*n));
  t = tos;
  if (!p) {
    if (segtable && (nseg != a->seg) && (nproc == 1) && (seg[nseg].use <= 0)) {
      u8 *s = &mem[segtable + 6*nseg];
      u16 u = getle16(s);
      if (u) {
        u16 b = getle16(s + 2);
        u16 l = getle16(s + 4);
        t -= l;
        unitread(u, t, l, b);
        if (*ioresult) execerr("IO error: %d", *ioresult);
        listsegment(nseg, &mem[t], l, 0);
      }
    }
    if ((nseg >= NBSEG) || (nproc == 0) || (nproc > seg[nseg].proccnt))
      execerr("attempt to call n%s procedure %u in segment %u",
              "on-existant", nproc, nseg);
    p = &seg[nseg].proc[nproc];
    seg[nseg].use += 1;
  }
  t -= p->psize + p->dsize;
  if (a) move_mem(tos, t, p->psize);
  n->par = t;
  tos = t;
  n->seg = nseg;
  n->proc = p;
  n->base = (p->lexlev <= 0) ? n : a->base;
  n->dyn = a;
  while (a && (p->lexlev <= a->proc->lexlev)) a = a->stat;
  n->stat = a;
  deb("proc %2d/%3d (base ", nseg, nproc);
  if (n->base->seg < NBSEG) {
    deb("%2d/%3d", n->base->seg, n->base->proc - &seg[n->base->seg].proc[0]);
  } else {
    deb("init");
  }
  deb("; stat ");
  if (!n->stat) {
    deb("none");
  } else if (n->stat->seg < NBSEG) {
    deb("%2d/%3d", n->stat->seg, n->stat->proc - &seg[n->stat->seg].proc[0]);
  } else {
    deb("init");
  }
  deb("; par/dat: %d/%d)\n", p->psize, p->dsize);
  ar = n;
  if (n->proc->jtab[0] != nproc)
    execerr("attempt to call n%s procedure %u in segment %u",
        "ative", nproc, nseg);
  return n->proc->ienter;
}

static void freear(unsigned rcount)
{
  activrec *a = ar;
  u32 t = a->par + 2*a->proc->psize + a->proc->dsize - 2*rcount;
  deb("tos %06lx, par %06lx, par/dat: %d/%d, r:%d; tnew: %06lx\n",
    tos, a->par, a->proc->psize, a->proc->dsize, rcount, t);
  move_mem(tos, t, 2*rcount);
  tos = t;
  ar = a->dyn;
  if (ar)
    seg[a->seg].use -= 1;
  a->dyn = unusedar;
  unusedar = a;
}

static activrec* imed(unsigned i)
{
  activrec *a = ar;
  while (i && a) {
    a = a->stat;
    i -= 1;
  }
  if (!a) execerr("static link chain exceeded");
  return a;
}

#define SETSIZ 256
static u16 set0[SETSIZ], set1[SETSIZ];

static void popset(u16 *t)
{
  u16 l = pop_int_check();
  if (l > SETSIZ) execerr("set size overflow");
  check_stack(2*l);
  memcpy(t, &mem[tos], 2*l);
  memset(&t[l], 0, (SETSIZ-l)*2);
  tos += 2*l;
}

static void pushset0(void)
{
  int i = SETSIZ-1;
  u16 l;
  while ((i >= 0) && (set0[i] == 0)) i -= 1;
  l = i+1;
  while (i >= 0) push_int(set0[i--]);
  push_int_check(l);
}

static void buildset(u16 a, u16 b)
{
  u16 *s = &set0[0];
  u16 v = 0;
  if (a <= b) {
    if ((a >= SETSIZ*16) || (b >= SETSIZ*16)) execerr("set range overrun");
    b -= a;
    while (a > 15) { *s++ = 0; a -= 16; }
    b += a;
    v = (u16) - (1 << a);
    while (b > 15) { *s++ = v; b -= 16; v = (u16)-1; }
    v &= (2 << b) - 1;
  }
  while (s < &set0[SETSIZ]) { *s++ = v; v = 0; }
}

static int pstrcmp(void)
{
  u8 *a, *b;
  u8 l;
  int r;
  b = &mem[pop_ref_check()];
  a = &mem[pop_ref_check()];
  l = *a; if (l > *b) l = *b;
  r = memcmp(&a[1], &b[1], l);
  if (r == 0) r = (int)*a - (int)*b;
  return r;
}

static int pmemcmp(int len)
{
  u8 *a, *b;
  b = &mem[pop_ref_check()];
  a = &mem[pop_ref_check()];
  return memcmp(a, b, len);
}

static u8 alphanum(u8 c)
{
  if (c >= 'A') {
    if ((c <= 'Z') || (c == '_')) return c;
    if ((c >= 'a') && (c <= 'z')) return c + 'A' - 'a';
  } else if ((c >= '0') && (c <= '9')) return c;
  return 0;
}

const static struct {
  u8 id[8];
  u8 sy;
  u8 op;
} idstab[] = {
  {"PROCEDUR", 31, 15},
  {"GOTO    ", 26, 15},
  {"TO      ",  7, 15},
  {"DOWNTO  ",  8, 15},
  {"NOT     ", 38, 15},
  {"SEGMENT ", 33, 15},
  {"USES    ", 49, 15},
  {"CASE    ", 21, 15},
  {"FILE    ", 46, 15},
  {"INTERFAC", 51, 15},
  {"OR      ", 40,  7},
  {"RECORD  ", 45, 15},
  {"SET     ", 42, 15},
  {"UNIT    ", 50, 15},
  {"WHILE   ", 23, 15},
  {"ARRAY   ", 44, 15},
  {"DIV     ", 39,  3},
  {"END     ",  9, 15},
  {"FORWARD ", 34, 15},
  {"IMPLEMEN", 52, 15},
  {"MOD     ", 39,  4},
  {"OF      ", 11, 15},
  {"PACKED  ", 43, 15},
  {"PROGRAM ", 33, 15},
  {"REPEAT  ", 22, 15},
  {"SEPARATE", 54, 15},
  {"THEN    ", 12, 15},
  {"TYPE    ", 29, 15},
  {"UNTIL   ", 10, 15},
  {"VAR     ", 30, 15},
  {"WITH    ", 25, 15},
  {"AND     ", 39,  2},
  {"BEGIN   ", 19, 15},
  {"CONST   ", 28, 15},
  {"DO      ",  6, 15},
  {"ELSE    ", 13, 15},
  {"EXTERNAL", 53, 15},
  {"FOR     ", 24, 15},
  {"FUNCTION", 32, 15},
  {"IF      ", 20, 15},
  {"IN      ", 41, 14},
  {"LABEL   ", 27, 15}
};
#define IDSTABSIZE (sizeof(idstab) / sizeof(idstab[0]))

static void idsearch(u32 s, u32 b)
{
  u16 symcursor = load_int(s);
  u16 sy, op;
  u8 *id = &mem[s+6];
  u8 *bufp = &mem[b];
  u8 c;
  int k, i = 0;
  while (c = alphanum(bufp[symcursor])) {
    symcursor += 1;
    if (i < 8) id[i++] = c;
  }
  store_int(s, symcursor-1);
  while (i < 8) id[i++] = ' ';
  i = 1;
  while ((i <= IDSTABSIZE) && (k = memcmp(&id[0], &idstab[i-1].id[0], 8)))
    i = (i << 1) + (k < 0 ? 0 : 1);
  if (i <= IDSTABSIZE) {
    sy = idstab[i-1].sy;
    op = idstab[i-1].op;
  } else {
    sy = 0;
    op = 15;
  }
  store_int(s+2, sy);
  store_int(s+4, op);
  deb("IDSEARCH(%06lx): SYMCURSOR=%04x, SY=%2d, OP=%2d, ID='%.8s'\n",
    b, symcursor, sy, op, &mem[s+6]);
}

static int treesearch(u32 i, u32 o, u32 a)
{
  u8 *ap = &mem[a];
  int k;
  u32 n;
  do {
    k = memcmp(ap, &mem[i], 8);
    n = i;
    i = load_any(i + 8 + (k < 0 ? 2 : 0));
  } while ((k != 0) && (i != PNIL));
  store_any(o, n);
  deb("TREESEARCH(%06lx, '%.8s'): %06lx, %d\n", i, ap, n, k);
  return k;
}

static u8 *callproc(u8 *ipc, unsigned nseg, unsigned nproc)
{
  if (apple) {
    switch (nseg) {
    case 22: /* APPLESTU */
      switch (nproc) {
      case 2: /* FUNCTION PADDLE(SELECT: INTEGER): INTEGER; */
        break;
      case 3: /* FUNCTION BUTTON(SELECT: INTEGER): BOOLEAN; */
        break;
      case 4: /* PROCEDURE TTLOUT(SELECT: INTEGER; DATA: BOOLEAN); */
        break;
      case 5: { /* FUNCTION KEYPRESS: BOOLEAN; */
          struct pollfd pfd;
          pop_any_check();
          pop_any_check();
          pfd.fd = 0;
          pfd.events = POLLIN;
          push_int(poll(&pfd, 1, 0) != 0);
        } return ipc;
      case 6: /* FUNCTION RANDOM: INTEGER; */
        break;
      case 7: /* PROCEDURE RANDOMIZE; */
        break;
      case 8: /* PROCEDURE NOTE(PITCH,DURATION: INTEGER); */
        break;
      }
      break;
    case 30: /* LONGINTI */
      break;
    case 31: /* TURTLEGR */
      break;
    }
  }
  ar->ipc = ipc;
  return allocar(nseg, nproc, NULL);
}

static u8 *retproc(unsigned rcount)
{
  freear(rcount);
  return ar ? ar->ipc : 0;
}

static u8 *exitproc(unsigned nseg, unsigned nproc)
{
  activrec *a = ar;
  do {
    a->ipc = a->proc->iexit;
  } while (((a->proc != &seg[nseg].proc[nproc]) || (nseg != a->seg))
        && (a = a->dyn));
  return ar->ipc;
}

/* signed byte: */
#define codesb() ((s8)(*ipc++))

/* big value (0..127 one byte, 128.. two bytes): */
#define codeb() ( { u8 b1 = *ipc++; (0 > (s8)b1) ? \
                        (((u16)(b1 & 0x7F)) << 8) + (u16)*ipc++ : (u16)b1; } )

/* signed word: */
#define codew() ( { u16 w = getle16(ipc); ipc += 2; w; } )

#define c04(n) case n: case n+1: case n+2: case n+3
#define c08(n) c04(n): c04(n+4)
#define c16(n) c08(n): c08(n+8)

#define ujp() ( { op = *ipc++; \
                if (0 > (s8)op) ipc = relpos16(ar->proc->jtab + (s8)op); \
                           else ipc += op; } )

static real rlog10;

#if DEBUG
#define OP_NO 0
#define OP_DB 1
#define OP_UB 2
#define OP_SB 3
#define OP_WD 4
#define OP_BG 5
#define OP_PS 6
#define OP_JP 7
#define OP_WA 8
#define OP_NI 9
#define OP_NN 10
const static struct debopinf {
  unsigned op1 : 4;
  unsigned op2 : 4;
  char name[3];
} debopinf[216-128] = {
  {OP_NO, OP_NO, "ABI"},
  {OP_NO, OP_NO, "ABR"},
  {OP_NO, OP_NO, "ADI"},
  {OP_NO, OP_NO, "ADR"},
  {OP_NO, OP_NO, "AND"},
  {OP_NO, OP_NO, "DIF"},
  {OP_NO, OP_NO, "DVI"},
  {OP_NO, OP_NO, "DVR"},
  {OP_NO, OP_NO, "CHK"},
  {OP_NO, OP_NO, "FLO"},
  {OP_NO, OP_NO, "FLT"},
  {OP_NO, OP_NO, "INN"},
  {OP_NO, OP_NO, "INT"},
  {OP_NO, OP_NO, "LOR"},
  {OP_NO, OP_NO, "MOD"},
  {OP_NO, OP_NO, "MPI"},
  {OP_NO, OP_NO, "MPR"},
  {OP_NO, OP_NO, "NGI"},
  {OP_NO, OP_NO, "NGR"},
  {OP_NO, OP_NO, "NOT"},
  {OP_NO, OP_NO, "SRS"},
  {OP_NO, OP_NO, "SBI"},
  {OP_NO, OP_NO, "SBR"},
  {OP_NO, OP_NO, "SGS"},
  {OP_NO, OP_NO, "SQI"},
  {OP_NO, OP_NO, "SQR"},
  {OP_NO, OP_NO, "STO"},
  {OP_NO, OP_NO, "IXS"},
  {OP_NO, OP_NO, "UNI"},
  {OP_NO, OP_NO, "S2P"},
  {OP_UB, OP_NO, "CSP"},
  {OP_NO, OP_NN, "LDC"},
  {OP_UB, OP_NO, "ADJ"},
  {OP_JP, OP_NO, "FJP"},
  {OP_BG, OP_NO, "INC"},
  {OP_BG, OP_NO, "IND"},
  {OP_BG, OP_NO, "IXA"},
  {OP_BG, OP_NO, "LAO"},
  {OP_PS, OP_NO, "LSA"},
  {OP_NO, OP_NO, "LAE"},
  {OP_BG, OP_NO, "MOV"},
  {OP_BG, OP_NO, "LDO"},
  {OP_UB, OP_NO, "SAS"},
  {OP_BG, OP_NO, "SRO"},
  {OP_WA, OP_WA, "XJP"},
  {OP_NO, OP_NO, "RNP"},
  {OP_UB, OP_NO, "CIP"},
  {OP_DB, OP_NO, "EQU"},
  {OP_DB, OP_NO, "GEQ"},
  {OP_DB, OP_NO, "GTR"},
  {OP_DB, OP_BG, "LDA"},
  {OP_UB, OP_WA, "LDC"}, /* block */
  {OP_DB, OP_NO, "LEQ"},
  {OP_DB, OP_NO, "LES"},
  {OP_DB, OP_BG, "LOD"},
  {OP_DB, OP_NO, "NEQ"},
  {OP_DB, OP_BG, "STR"},
  {OP_JP, OP_NO, "UJP"},
  {OP_NO, OP_NO, "LDP"},
  {OP_NO, OP_NO, "STP"},
  {OP_UB, OP_NO, "LDM"},
  {OP_UB, OP_NO, "STM"},
  {OP_NO, OP_NO, "LDB"},
  {OP_NO, OP_NO, "STB"},
  {OP_DB, OP_DB, "IXP"},
  {OP_NO, OP_NO, "RBP"},
  {OP_UB, OP_NO, "CBP"},
  {OP_NO, OP_NI, "EQU"},
  {OP_NO, OP_NI, "GEQ"},
  {OP_NO, OP_NI, "GTR"},
  {OP_BG, OP_NO, "LLA"},
  {OP_WD, OP_NI, "LDC"},
  {OP_NO, OP_NI, "LEQ"},
  {OP_NO, OP_NI, "LES"},
  {OP_BG, OP_NO, "LDL"},
  {OP_NO, OP_NI, "NEQ"},
  {OP_BG, OP_NO, "STL"},
  {OP_DB, OP_UB, "CXP"},
  {OP_UB, OP_NO, "CLP"},
  {OP_UB, OP_NO, "CGP"},
  {OP_PS, OP_NO, "LPA"},
  {OP_NO, OP_NO, "IXB"},
  {OP_NO, OP_NO, "BYT"},
  {OP_JP, OP_NO, "EFJ"},
  {OP_JP, OP_NO, "NFJ"},
  {OP_NO, OP_NO, "BPT"},
  {OP_NO, OP_NO, "XIT"},
  {OP_NO, OP_NO, "NOP"}
};

static const struct debopinf *getopinf(u8 op)
{
  static struct {
    unsigned op1 : 4;
    unsigned op2 : 4;
    char name[8];
  } v;
  if (((s8)op) >= 0) {
    sprintf(&v.name[0], "SLDC%d", op);
    v.op1 = OP_NO; v.op2 = OP_NO;
  } else if (op < 216) {
    *(struct debopinf *)&v = debopinf[op - 128];
    switch (v.op2) {
    case OP_NI: v.name[3] = 'I'; v.name[4] = 0; break;
    case OP_NN: v.name[3] = 'N'; v.name[4] = 0; break;
       default: v.name[3] = 0; break;
    }
  } else if (op < 232) {
    sprintf(&v.name[0], "SLDL%d", op - 216 + 1);
    v.op1 = OP_NO; v.op2 = OP_NO;
  } else if (op < 248) {
    sprintf(&v.name[0], "SLDO%d", op - 232 + 1);
    v.op1 = OP_NO; v.op2 = OP_NO;
  } else {
    sprintf(&v.name[0], "SIND%d", op - 248);
    v.op1 = OP_NO; v.op2 = OP_NO;
  }
  return &v;
}

static int putopinf(u8 op, u8 *c, u8 *ipc, char **t)
{
  u8 b = c[0];
  int r = 0;
  switch (op) {
  case OP_NO:
  case OP_NI:
  case OP_NN:
    break;
  case OP_DB:
  case OP_SB:
    *t += sprintf(*t, " %d", (s8)b);
    r = 1;
    break;
  case OP_UB:
    *t += sprintf(*t, " %u", b);
    r = 1;
    break;
  case OP_WA:
    if (((unsigned)c) & 1) {
      c += 1;
      b = c[0];
      r += 1;
    }
  case OP_WD:
    *t += sprintf(*t, " %d", (s16)((c[1] << 8) + b));
    r += 2;
    break;
  case OP_BG:
    *t += sprintf(*t, " %u", (0 > (s8)b) ?
        (((u16)(b & 0x7F)) << 8) + (u16)c[1] : (u16)b);
    r = (0 > (s8)b) ? 2 : 1;
    break;
  case OP_PS:
    *t += sprintf(*t, " %u ", b);
    r = 1;
    while (r <= b) {
      u8 a = c[r];
      *t += sprintf(*t, "%c", ((a < ' ') || (a > 126)) ? '.' : a);
      r += 1;
    }
    break;
  case OP_JP:
    *t += sprintf(*t, " %06lx", ((0 > (s8)b) ?
        relpos16(ar->proc->jtab + (s8)b) : ipc + 2 + b) - mem);
    r = 1;
    break;
  }
  return r;
}
#endif

static void execpcod(u8 *ipc)
{
  do {
    u8 op;
#if DEBUG
    if (opt_debug) {
      char txt[512];
      char *t = &txt[0];
      char *x;
      int i, j;
      const struct debopinf *d = getopinf(ipc[0]);
      if (ar->seg < NBSEG) {
        t = t + sprintf(t, "%2d/%3d",
          ar->seg, ar->proc - &seg[ar->seg].proc[0]);
      } else {
        t = t + sprintf(t, "  init");
      }
      x = t + sprintf(t, "/%06x", ipc - mem);
      t = x + sprintf(x, ":%17s", &d->name[0]);
      i = 1 + putopinf(d->op1, &ipc[1], ipc, &t);
      i += putopinf(d->op2, &ipc[i], ipc, &t);
      if (i > 6) i = 6;
      for (j = 0; x += 2, j < i; j++) snprintf(x, 3, "%02x", ipc[j]);
      *x = ' ';
      i = 42 + &txt[0] - t;
      if (i < 0) t += sprintf(t, "\n%43c", ' ');
        else do *t++ = ' '; while (--i >= 0);
      t += sprintf(t, "%06lx:", tos);
      for (j = 0; j < 6; j += 2) {
        i = bank[(tos+j) >> 1];
        if (i) t += sprintf(t, " %02x", i);
          else t += sprintf(t, "   ");
        t += sprintf(t, "%04x", *(u16*)(mem+tos+j));
      }
      t += sprintf(t, "...\n");
      fputs(&txt[0], stderr);
    }
#endif
    op = *ipc++;
    if (0 <= (s8)op) {
      push_int_check(op);
    } else switch (op) {
/*ABI*/ case 128: { s16 w = pop_int_check();
                    if (w < 0) w = -w;
                    push_int(w);
                  } break;
/*ABR*/ case 129: { real r = pop_real_check();
                    if (r < 0) r = -r;
                    push_real(r);
                  } break;
/*ADI*/ case 130: { if (is_ref(tos)) {
                      u32 a = pop_ref_check();
                      push_ref(a + (s16)pop_int_check());
                    } else {
                      s16 w = pop_int_check();
                      if (is_ref(tos)) push_ref(pop_ref_check() + w);
                        else push_int(w + (s16)pop_int_check());
                    }
                  } break;
/*ADR*/ case 131: { push_real(pop_real_check() + pop_real_check());
                  } break;
/*AND*/ case 132: { push_int(pop_int_check() & pop_int_check());
                  } break;
/*DIF*/ case 133: { s16 w;
                    popset(&set1[0]); popset(&set0[0]); w = SETSIZ-1;
                    do set0[w] &= ~set1[w]; while (--w >= 0);
                    pushset0();
                  } break;
/*DVI*/ case 134: { s16 w = pop_int_check();
                    push_int((s16)pop_int_check() / w); /* Wirth? */
                  } break;
/*DVR*/ case 135: { real r = pop_real_check();
                    push_real(pop_real_check() / r);
                  } break;
/*CHK*/ case 136: { s16 w = load_int_check(tos + 4);
                    if ((w > (s16)pop_int_check())
                     || (w < (s16)pop_int_check())) execerr("CHK failed");
                  } break;
/*FLO*/ case 137: { s16 w = pop_int_check();
                    push_real((s16)pop_int_check());
                    push_int_check(w);
                  } break;
/*FLT*/ case 138: { push_real_check((s16)pop_int_check());
                  } break;
/*INN*/ case 139: { u16 w;
                    popset(&set0[0]);
                    w = pop_int_check();
                    if (w >= SETSIZ*16) push_int(0);
                    else push_int((set0[w/16] >> w%16) & 1);
                  } break;
/*INT*/ case 140: { s16 w;
                    popset(&set1[0]); popset(&set0[0]); w = SETSIZ-1;
                    do set0[w] &= set1[w]; while (--w >= 0);
                    pushset0();
                  } break;
/*LOR*/ case 141: { push_int(pop_int_check() | pop_int_check());
                  } break;
/*MOD*/ case 142: { s16 w = pop_int_check();
                    push_int((s16)pop_int_check() % w); /* Wirth? */
                  } break;
/*MPI*/ case 143: { push_int((s16)pop_int_check() * (s16)pop_int_check());
                  } break;
/*MPR*/ case 144: { push_real(pop_real_check() * pop_real_check());
                  } break;
/*NGI*/ case 145: { push_int(-(s16)pop_int_check());
                  } break;
/*NGR*/ case 146: { push_real(-pop_real_check());
                  } break;
/*NOT*/ case 147: { push_int(~pop_int_check());
                  } break;
/*SRS*/ case 148: { s16 w = pop_int_check();
                    buildset(pop_int_check(), w);
                    pushset0();
                  } break;
/*SBI*/ case 149: { if (is_ref(tos)) {
                      u32 a = pop_ref_check();
                      push_int(pop_ref_check() - a);
                    } else {
                      s16 w = pop_int_check();
                      if (is_ref(tos)) push_ref(pop_ref_check() - w);
                        else push_int(pop_int_check() - w);
                    }
                  } break;
/*SBR*/ case 150: { real r = pop_real_check();
                    push_real(pop_real_check() - r);
                  } break;
/*SGS*/ case 151: { s16 w = pop_int_check();
                    buildset(w, w);
                    pushset0();
                  } break;
/*SQI*/ case 152: { s16 w = pop_int_check();
                    push_int(w*w);
                  } break;
/*SQR*/ case 153: { real r = pop_real_check();
                    push_real(r*r);
                  } break;
/*STO*/ case 154: { u32 w = pop_any_check();
                    store_any_check(pop_ref_check(), w);
                  } break;
/*IXS*/ case 155: { s16 w = pop_int_check();
                    u32 r = pop_ref_check();
                    if ((w < 1) || (w > mem[r]))
                      execerr("string index range fault");
                    push_ref(r + w);
                  } break;
/*UNI*/ case 156: { s16 w;
                    popset(&set1[0]); popset(&set0[0]); w = SETSIZ-1;
                    do set0[w] |= set1[w]; while (--w >= 0);
                    pushset0();
                  } break;
/*S2P   case 157:*/
/*CSP*/ case 158: op = *ipc++;
                  switch (op) {
/*IOCHK*/case  0: { if (*ioresult) execerr("IO error: %d", *ioresult);
                  } break;
/*NEW*/  case  1: { s16 w = pop_int_check();
                    store_ref_check(pop_ref_check(), new_heap(2*w));
                  } break;
/*MVL*/  case  2: { s16 cnt, ofs;
                    u32 dst, src;
                    cnt = pop_int_check();
                    ofs = pop_int_check();
                    dst = pop_ref_check() + ofs;
                    ofs = pop_int_check();
                    src = pop_ref_check() + ofs;
                    move_left(src, dst, cnt);
                  } break;
/*MVR*/  case  3: { s16 cnt, ofs;
                    u32 dst, src;
                    cnt = pop_int_check();
                    ofs = pop_int_check();
                    dst = pop_ref_check() + ofs;
                    ofs = pop_int_check();
                    src = pop_ref_check() + ofs;
                    move_right(src, dst, cnt);
                  } break;
/*EXIT*/ case  4: { s16 w = pop_int_check();
                    ipc = exitproc(pop_int_check(), w);
                  } break;
/*UREAD*/case  5: { u16 b; s16 s; u32 a; int i; u16 u;
                    pop_stack(2);
                    b = pop_int_check();
                    s = pop_int_check();
                    i = pop_int_check();
                    a = pop_ref_check() + i;
                    u = pop_int_check();
                    unitread(u, a, s, b);
                  } break;
/*UWRIT*/case  6: { u16 b; s16 s; u32 a; int i; u16 u;
                    pop_stack(2);
                    b = pop_int_check();
                    s = pop_int_check();
                    i = pop_int_check();
                    a = pop_ref_check() + i;
                    u = pop_int_check();
                    unitwrite(u, a, s, b);
                  } break;
/*IDS*/  case  7: { u32 b; u32 s;
                    b = pop_ref_check();
                    s = pop_ref_check();
                    idsearch(s, b);
                  } break;
/*TRS*/  case  8: { u32 a; u32 o; u32 i;
                    a = pop_ref_check();
                    o = pop_ref_check();
                    i = pop_ref_check();
                    push_int(treesearch(i, o, a));
                  } break;
/*TIM*/  case  9: { store_int_check(pop_ref_check(), 0);
                    store_int_check(pop_ref_check(), 0);
                  } break;
/*FLC*/  case 10: { u16 size; int i;
                    op = pop_int_check();
                    size = pop_int_check();
                    i = pop_int_check();
                    fill_mem(pop_ref_check() + i, size, op);
                  } break;
/*SCN*/  case 11: { s16 w; u32 a; int comp; int i;
                    pop_stack(2);
                    w = pop_int_check();
                    a = pop_ref_check() + w;
                    op = pop_int_check();
                    comp = pop_int_check();
                    w = pop_int_check();
                    i = 0;
                    while ((i != w)
                        && (comp ? (mem[a+i] == op) : (mem[a+i] != op)))
                      i += (w < 0) ? -1 : 1;
                    push_int(i);
                  } break;
/*GETSE*/case 21: { u16 w = pop_int_check();
                    if (w >= NBSEG)
                      execerr("segment not available: %d", w);
                  } break;
/*RELSE*/case 22: { pop_stack(2);
                  } break;
/*TNC*/  case 23: { push_int(truncf(pop_real_check())); /* Wirth? */
                  } break;
/*RND*/  case 24: { push_int(roundf(pop_real_check())); /* Wirth? */
                  } break;
/*SIN*/  case 25: { push_real(sinf(pop_real_check()));
                  } break;
/*COS*/  case 26: { push_real(cosf(pop_real_check()));
                  } break;
/*LOG*/  case 27: { push_real(logf(pop_real_check()) * rlog10);
                  } break;
/*ATAN*/ case 28: { push_real(atanf(pop_real_check()));
                  } break;
/*LN*/   case 29: { push_real(logf(pop_real_check()));
                  } break;
/*EXP*/  case 30: { push_real(expf(pop_real_check()));
                  } break;
/*SQT*/  case 31: { push_real(sqrtf(pop_real_check()));
                  } break;
/*MRK*/  case 32: { store_ref_check(pop_ref_check(), toh);
                    deb("heap mark: %06x\n", toh);
                  } break;
/*RLS*/  case 33: { u32 a = load_ref_check(pop_ref_check());
                    deb("heap release %06x -> %06x\n", toh, a);
                    if (a > toh) execerr("heap release overflow");
                    toh = a;
                  } break;
/*IORES*/case 34: { push_int_check(*ioresult);
                  } break;
/*UBUSY*/case 35: { pop_stack(2);
                    push_int(0);
                  } break;
/*POT*/  case 36: { u16 w = pop_int_check();
                    if (w > 38)
                      execerr("power range fault");
                    push_real_check(exp10f(w));
                  } break;
/*UWAIT*/case 37: { pop_stack(2);
                  } break;
/*UCLEA*/case 38: { u16 w = pop_int_check();
                    *ioresult = pcerrno((w >= NBUNIT) ? ENXIO :
                      ((unitr[w] < 0) && (unitw[w] < 0)) ? ENODEV : 0);
                    if (unitt[w] < 0) unitt[w] = -unitt[w];
                  } break;
/*HALT*/ case 39: execerr("HALT");
/*MEMAV*/case 40: { u32 u = (tos - toh) / 2;
                    if (u > 32767) u = 32767;
                    push_int_check(u);
                  } break;
     case SETIOR: { if (ioresult) execerr("multiple ioresult");
                    ioresult = &mem[pop_ref_check()];
                  } break;
     case SETSEG: { if (segtable) execerr("multiple segtable");
                    segtable = pop_ref_check();
                    fill_mem(segtable, 32*3*2, 0);
                  } break;
     case SETDIR: { if (globdirp) execerr("multiple globdirp");
                    globdirp = pop_ref_check();
                  } break;
     case SETCRT: { if (crtctrl) execerr("multiple crtctrl");
                    crtctrl = &mem[pop_ref_check()];
                  } break;
         default: execerr("unknown CSP %02x", op);
                  } break;
/*LDCN*/case 159: { push_int(PNIL);
                  } break;
/*ADJ*/ case 160: { popset(&set0[0]);
                    op = *ipc++ * 2;
                    tos -= op;
                    memcpy(&mem[tos], &set0, op);
                    memset(&bank[tos >> 1], 0, op >> 1);
                  } break;
/*FJP*/ case 161: { if (!(pop_int_check() & 1)) ujp();
                      else ipc += 1;
                  } break;
/*INC*/ case 162: { push_ref(pop_ref_check() + 2 * codeb());
                  } break;
/*IND*/ case 163: { push_any(load_any(pop_ref_check() + 2 * codeb()));
                  } break;
/*IXA*/ case 164: { s16 w = pop_int_check();
                    push_ref(pop_ref_check() + 2 * codeb() * w);
                  } break;
/*LAO*/ case 165: { push_ref_check(ar->base->par + 2 * (codeb()-1));
                  } break;
/*LSA*/ case 166: { push_ref_check(ipc - mem);
                    op = *ipc++;
                    ipc += op;
                  } break;
/*LAE   case 167:*/
/*MOV*/ case 168: { u32 src = pop_ref_check();
                    u32 dst = pop_ref_check();
                    move_mem(src, dst, codeb() * 2);
                  } break;
/*LDO*/ case 169: { push_any_check(load_any(ar->base->par + 2 * (codeb()-1)));
                  } break;
/*SAS*/ case 170: { u32 src = pop_any_check();
                    u32 dst = pop_ref_check();
                    op = *ipc++;
                    if (src < (1 << 16)) {
                      if (op < 1) execerr("string assign overrun");
                      mem[dst] = 1; mem[dst + 1] = src;
                      bank[dst >> 1] = 0; bank[(dst+1) >> 1] = 0;
                    } else {
                      if (op < mem[src]) execerr("string assign overrun");
                      move_mem(src, dst, mem[src] + 1);
{ unsigned char tmp[257]; memset(&tmp[0], 0, 257);
  memcpy(&tmp[0], &mem[dst], mem[dst] + 1);
  deb("SAS: %u \"%s\"\n", tmp[0], &tmp[1]); }
                    }
                  } break;
/*SRO*/ case 171: { store_any(ar->base->par + 2 * (codeb()-1), pop_any_check());
                  } break;
/*XJP*/ case 172: { s16 m, w = pop_int_check();
                    ipc = (u8*)align(((u32)ipc) + 1);
                    m = codew();
                    if ((w <= codew()) && (w >= m))
                      ipc = relpos16((u8*)&((u16*)ipc)[w - m + 1]);
                  } break;
/*RNP*/ case 173: { ipc = retproc(*ipc);
                  } break;
/*CIP*/ case 174: { ipc = callproc(ipc+1, ar->seg, ipc[0]);
                  } break;
/*EQU*/ case 175: switch (*ipc++) {
         case  2: { push_int(pop_real_check() == pop_real_check());
                  } break;
         case  4: { push_int(pstrcmp() == 0);
                  } break;
         case  6: { push_int((pop_int_check() + pop_int_check() + 1) & 1);
                  } break;
         case  8: { popset(&set0[0]); popset(&set1[0]);
                    push_int(memcmp(&set0, &set1, sizeof(set0)) == 0);
                  } break;
         case 10: { push_int(pmemcmp(*ipc++) == 0);
                  } break;
         case 12: { push_int(pmemcmp(*ipc++ * 2) == 0);
                  } break;
                  } break;
/*GEQ*/ case 176: switch (*ipc++) {
         case  2: { real r = pop_real_check();
                    push_int(pop_real_check() >= r);
                  } break;
         case  4: { push_int(pstrcmp() >= 0);
                  } break;
         case  6: { u16 w = pop_int_check() & 1;
                    push_int((pop_int_check() & 1) >= w);
                  } break;
         case  8: { int i = SETSIZ-1;
                    popset(&set0[0]); popset(&set1[0]);
                    while ((i >= 0) && !(set0[i] & ~set1[i])) i -= 1;
                    push_int(i < 0);
                  } break;
         case 10: { push_int(pmemcmp(*ipc++) >= 0);
                  } break;
                  } break;
/*GTR*/ case 177: switch (*ipc++) {
         case  2: { real r = pop_real_check();
                    push_int(pop_real_check() > r);
                  } break;
         case  4: { push_int(pstrcmp() > 0);
                  } break;
         case  6: { u16 w = pop_int_check() & 1;
                    push_int((pop_int_check() & 1) > w);
                  } break;
         case 10: { push_int(pmemcmp(*ipc++) > 0);
                  } break;
                  } break;
/*LDA*/ case 178: { op = *ipc++;
                    push_ref_check(imed(op)->par + 2 * (codeb()-1));
                  } break;
/*LDC*/ case 179: { op = *ipc;
                    ipc = (u8*)align(((u32)ipc) + 2);
                    while (op) {
                      push_int_check(*(u16*)ipc);
                      ipc += 2;
                      op -= 1;
                    }
                  } break;
/*LEQ*/ case 180: switch (*ipc++) {
         case  2: { real r = pop_real_check();
                    push_int(pop_real_check() <= r);
                  } break;
         case  4: { push_int(pstrcmp() <= 0);
                  } break;
         case  6: { u16 w = pop_int_check() & 1;
                    push_int((pop_int_check() & 1) <= w);
                  } break;
         case  8: { int i = SETSIZ-1;
                    popset(&set1[0]); popset(&set0[0]);
                    while ((i >= 0) && !(set0[i] & ~set1[i])) i -= 1;
                    push_int(i < 0);
                  } break;
         case 10: { push_int(pmemcmp(*ipc++) <= 0);
                  } break;
                  } break;
/*LES*/ case 181: switch (*ipc++) {
         case  2: { real r = pop_real_check();
                    push_int(pop_real_check() < r);
                  } break;
         case  4: { push_int(pstrcmp() < 0);
                  } break;
         case  6: { u16 w = pop_int_check() & 1;
                    push_int((pop_int_check() & 1) < w);
                  } break;
         case 10: { push_int(pmemcmp(*ipc++) < 0);
                  } break;
                  } break;
/*LOD*/ case 182: { op = *ipc++;
                    push_any_check(load_any(imed(op)->par + 2 * (codeb()-1)));
                  } break;
/*NEQ*/ case 183: switch (*ipc++) {
         case  2: { push_int(pop_real_check() != pop_real_check());
                  } break;
         case  4: { push_int(pstrcmp() != 0);
                  } break;
         case  6: { push_int((pop_int_check() + pop_int_check()) & 1);
                  } break;
         case  8: { popset(&set0[0]); popset(&set1[0]);
                    push_int(memcmp(&set0, &set1, sizeof(set0)) != 0);
                  } break;
         case 10: { push_int(pmemcmp(*ipc++) != 0);
                  } break;
         case 12: { push_int(pmemcmp(*ipc++ * 2) != 0);
                  } break;
                  } break;
/*STR*/ case 184: { op = *ipc++;
                    store_any(imed(op)->par + 2 * (codeb()-1), pop_any_check());
                  } break;
/*UJP*/ case 185: { ujp();
                  } break;
/*LDP*/ case 186: { u16 shft = pop_int_check();
                    u16 bits = pop_int_check();
                    if (shft + bits > 16)
                      execerr("packed field pointer exceeds word\n");
                    push_int((load_int_check(pop_ref_check()) >> shft)
                      & ((1 << bits) - 1));
                  } break;
/*STP*/ case 187: { u32 a;
                    u16 w = pop_int_check();
                    u16 shft = pop_int_check();
                    u16 bits = pop_int_check();
                    if (shft + bits > 16)
                      execerr("packed field pointer exceeds word\n");
                    a = pop_ref_check();
                    store_int_check(a,
                      (load_int(a) & ~(((1 << bits) - 1) << shft))
                        | ((w & ((1 << bits) - 1)) << shft));
                  } break;
/*LDM*/ case 188: { u32 a = pop_ref_check();
                    op = *ipc++;
                    tos -= 2*op;
                    move_mem(a, tos, 2*op);
                  } break;
/*STM*/ case 189: { u32 a = tos;
                    op = *ipc++;
                    tos += 2*op;
                    move_mem(a, pop_ref_check(), 2*op);
                  } break;
/*LDB*/ case 190: { s16 w = pop_int_check();
                    push_int(mem[pop_ref_check() + w]);
                  } break;
/*STB*/ case 191: { s16 w;
                    u32 a;
                    op = pop_int_check();
                    w = pop_int_check();
                    a = pop_ref_check() + w;
                    mem[a] = op;
                    bank[a >> 1] = 0;
                  } break;
/*IXP*/ case 192: { u32 a;
                    s16 w = pop_int_check();
                    op = *ipc++;
                    a = pop_ref_check();
                    if (w >= 0) { a += (w/op)*2; w = w%op; }
                      else { int i = (op-1-w)/op; a -= i*2; w += i*op; }
                    push_ref(a);
                    op = *ipc++;
                    push_int(op);
                    push_int_check(op * w);
                  } break;
/*RBP*/ case 193: { ipc = retproc(*ipc);
                  } break;
/*CBP*/ case 194: { ipc = callproc(ipc+1, ar->seg, ipc[0]);
                  } break;
/*EQUI*/case 195: { push_int(pop_any_check() == pop_any_check());
                  } break;
/*GEQI*/case 196: { s16 w = pop_int_check();
                    push_int(w <= (s16)pop_int_check());
                  } break;
/*GTRI*/case 197: { s16 w = pop_int_check();
                    push_int(w < (s16)pop_int_check());
                  } break;
/*LLA*/ case 198: { push_ref_check(ar->par + 2 * (codeb()-1));
                  } break;
/*LDCI*/case 199: { push_int(codew());
                  } break;
/*LEQI*/case 200: { s16 w = pop_int_check();
                    push_int(w >= (s16)pop_int_check());
                  } break;
/*LESI*/case 201: { s16 w = pop_int_check();
                    push_int(w > (s16)pop_int_check());
                  } break;
/*LDL*/ case 202: { push_any_check(load_any(ar->par + 2 * (codeb()-1)));
                  } break;
/*NEQI*/case 203: { push_int(pop_any_check() != pop_any_check());
                  } break;
/*STL*/ case 204: { store_any(ar->par + 2 * (codeb()-1), pop_any_check());
                  } break;
/*CXP*/ case 205: { ipc = callproc(ipc+2, ipc[0], ipc[1]);
                  } break;
/*CLP*/ case 206: { ipc = callproc(ipc+1, ar->seg, ipc[0]);
                  } break;
/*CGP*/ case 207: { ipc = callproc(ipc+1, ar->seg, ipc[0]);
                  } break;
/*LPA*/ case 208: { op = *ipc++;
                    push_ref_check(ipc - mem);
                    ipc += op;
                  } break;
/*IXB*/ case 209: { s16 w = pop_int_check();
                    push_ref(pop_ref_check() + w);
                  } break;
/*BYT*/ case 210: { check_ref(tos);
                  } break;
/*EFJ*/ case 211: { if (pop_any_check() != pop_any_check()) ujp();
                      else ipc += 1;
                  } break;
/*NFJ*/ case 212: { if (pop_any_check() == pop_any_check()) ujp();
                      else ipc += 1; 
                  } break;
/*BPT   case 213:*/
/*XIT*/ case 214: { ar = NULL;
                  } break;
/*NOP*/ case 215: break;
/*SLDL*/c16(216): { push_any_check(load_any(ar->par + 2 * (op - 216)));
                  } break;
/*SLDO*/c16(232): { push_any_check(load_any(ar->base->par + 2 * (op - 232)));
                  } break;
/*SIND*/c08(248): { push_any(load_any(pop_ref_check() + 2 * (op - 248)));
                  } break;
         default: execerr("unknown opcode: %d", op);
    }
  } while (ar);
}

int main(int argc, char *argv[])
{
  int i, e, j;
  unsigned s = NBSEG;
  procdescr p;
  const u8 *c;
  i = chkoptions(argc, argv);
  if (unitt[1] == 2) {
    initscr();
    noecho();
    halfdelay(8);
    keypad(stdscr, TRUE);
    scrollok(stdscr, TRUE);
    signal(SIGINT, (void *) (*signalhandler));
  }
  if (!(mem = malloc(MAX_BANK << 16))
   || !(bank = malloc(MAX_BANK << 15))) execerr("malloc failed");
  toh = 1 << 16;
  mem -= (1 << 16);
  bank -= (1 << 15);
  tos = (MAX_BANK + 1) << 16;
  fetchsegments(readcode(argv[i]));
  if (seg[0].code) {
    c = &code40[0];
    e = sizeof(code40);
    j = 3;
  } else {
    c = &code41[0];
    e = sizeof(code41);
    j = 1;
  }
  while (++i < argc) fetchsegments(readcode(argv[i]));
  if (!seg[4].code) {
    execerr("global initialisation code missing");
  }
  p.ienter = &mem[new_heap(align(e+2))];
  memcpy(p.ienter, c, e);
  p.jtab = p.ienter + e;
  p.jtab[0] = 1;
  p.iexit = p.jtab - j;
  p.psize = 0;
  p.dsize = 774;
  p.lexlev = -1;
  memset(&bank[1 << 15], 0, (toh - (1 << 16)) / 2 + 1);
  rlog10 = 1 / logf(10);
  ioresult = 0;
  crtctrl = 0;
  segtable = 0;
  globdirp = 0;
  deb("RUN s=%d, p=%p\n", s, p);
  execpcod(allocar(s, 1, &p));
  exiterr(0);
}
